#include <chrono>
#include <iostream>
#include <vector>
#include <string>
#include <algorithm>
#include <thread>
#include <mutex>
#include <condition_variable>

#define WORK_ITERATIONS 100

using namespace std;
using namespace std::chrono;


class Tool {
public:
    string mName;
    int mAvailableUnits;

    Tool(string name, int availableUnits) : mName(name), mAvailableUnits(availableUnits) {}
};

class Guy {
private:
    string mName;
    Tool &mTool1;
    Tool &mTool2;
    mutex &mBorrowOrReturnMutex;
    mutex &mStdoutMutex;
    condition_variable &mCondVar;

    void work() {
        // Heavy work of a guy, just a random delay...
        this_thread::sleep_for(chrono::milliseconds(rand() % 30));
    }

    void borrowTools(const vector<Tool*> &toolsToBorrow) {
        unique_lock<mutex> lock(mBorrowOrReturnMutex);
        mCondVar.wait(lock, [&toolsToBorrow]() {
            // 'borrowOrReturnMutex' is locked here.
            return all_of(
                    toolsToBorrow.cbegin(),
                    toolsToBorrow.cend(),
                    [](Tool *tool) { return tool->mAvailableUnits > 0; });
        });


        // If a thread is woken and the predicate is satisfied (all tools available),
        // then mutex 'borrowOrReturnMutex' is still locked and we can modify the tool numbers.

        lock_guard<mutex> outputGuard(mStdoutMutex);
        cout << "Guy " << mName << " borrowed: ";
        for (auto *tool : toolsToBorrow) {
            tool->mAvailableUnits--;
            cout << tool->mName << ", ";
        }
        cout << endl;

        // After the method is finished, mutex 'borrowOrReturnMutex' will be unlocked.
    }

    void returnTools(const vector<Tool*> &toolsToReturn) {
        {
            lock_guard<mutex> lockGuard(mBorrowOrReturnMutex);
            lock_guard<mutex> outputGuard(mStdoutMutex);

            // Mutex is locked, no other thread can return or borrow tools until it is unlocked.
            cout << "Guy " << mName << " returned: ";
            for (auto *tool : toolsToReturn) {
                tool->mAvailableUnits++;
                cout << tool->mName << ", ";
            }
            cout << endl;
        }

        // Notify that some tools are returned and try to awake all threads one by one.
        mCondVar.notify_all();
    }

public:
    Guy(string name, Tool &tool1, Tool &tool2, mutex &borrowOrReturnMutex, mutex &stdoutMutex, condition_variable &condVar)
        : mName(name), mTool1(tool1), mTool2(tool2),
          mBorrowOrReturnMutex(borrowOrReturnMutex), mStdoutMutex(stdoutMutex), mCondVar(condVar) {}

    void startGuy() {
        // Each guy is iteratively doing the job.
        for (int iter = 0; iter < WORK_ITERATIONS; ++iter) {

            // Work with one tool
            borrowTools({&mTool1});
            work();
            returnTools({&mTool1});

            // Work with two tools
            borrowTools({&mTool1, &mTool2});
            work();
            returnTools({&mTool1, &mTool2});
        }

        return;
    }
};

int main() {
    mutex stdoutMutex;  // Mutex used to avoid garbled stdout.
    mutex borrowOrReturnMutex;  // Mutex used for the following condition variable.
    condition_variable condVar; // Threads are notified if tools are returned.

    Tool hammer("hammer", 1);
    Tool screwDriver("screw driver", 1);
    Tool saw("saw", 1);

    Guy libor("Libor", hammer, screwDriver, borrowOrReturnMutex, stdoutMutex, condVar);
    Guy honza("Honza", screwDriver, saw, borrowOrReturnMutex, stdoutMutex, condVar);
    Guy premek("Premek", saw, hammer, borrowOrReturnMutex, stdoutMutex, condVar);

    vector<thread> guyThreads;
    guyThreads.emplace_back(&Guy::startGuy, libor);
    guyThreads.emplace_back(&Guy::startGuy, honza);
    guyThreads.emplace_back(&Guy::startGuy, premek);

    for (auto &guyThread: guyThreads) {
        guyThread.join();
    }

    return 0;
}
